! Copyright (C) 2021 National Center for Atmospheric Research
! SPDX-License-Identifier: Apache-2.0
!
!> \file
!> Tests for the photo_grid module

!> Test module for the grid_t type
program test_grid

  use musica_mpi,                      only : musica_mpi_init,                &
                                              musica_mpi_finalize

  implicit none

  call musica_mpi_init( )
  call test_grids( )
  call musica_mpi_finalize( )

contains

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

  subroutine test_grid_t(aGrid, num_cells, edges, midpoints, deltas)
    use musica_assert,    only : assert, almost_equal
    use musica_constants, only : dk => musica_dk
    use tuvx_grid,    only : grid_t

    integer :: i

    class(grid_t), pointer   :: aGrid
    integer :: num_cells
    real(dk) :: edges(:), midpoints(:), deltas(:)

    call assert( 412238776, all( aGrid%delta_ > 0._dk ) )

    call assert( 412238777, aGrid%ncells_ .eq. num_cells )
    call assert( 551632826, aGrid%size( ) .eq. num_cells )

    ! plus 1 here because there are always 2 more edges than cells
    call assert( 412238778, size( aGrid%edge_ ) .eq. num_cells + 1 )

    do i = 1, size(aGrid%edge_)
      call assert( 412238779, &
        almost_equal( edges(i), aGrid%edge_(i), 0.1d-5 ) )
    end do

    call assert( 412238780, size( aGrid%mid_ ) .eq. num_cells )
    do i = 1, size(aGrid%mid_)
      call assert( 412238781, &
        almost_equal( midpoints( i ), aGrid%mid_(i), 0.1d-5 ) )
    end do

    call assert( 412238782, size( aGrid%delta_ ) .eq. num_cells )

    do i = 1, size(aGrid%delta_)
      call assert( 412238783, &
        almost_equal( deltas(i), aGrid%delta_(i), 0.1d-5 ) )
    end do

    call aGrid%output( 6 )

  end subroutine test_grid_t

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

  !> Tests of the grid_t type
  subroutine test_grids( )

    use musica_assert,                 only : assert
    use musica_config,                 only : config_t
    use musica_string,                 only : string_t
    use musica_constants,              only : ik => musica_ik, dk => musica_dk
    use musica_mpi
    use tuvx_grid,                     only : grid_t
    use tuvx_grid_factory,             only : grid_type_name, grid_allocate
    use tuvx_grid_warehouse,           only : grid_warehouse_t

    !> local variables
    character(len=*), parameter :: config_flsp = 'test/data/grid.test.config.json'
    type(config_t) :: grid_tst_config
    type(grid_warehouse_t), pointer :: thewarehouse
    class(grid_t), pointer   :: aGrid
    integer :: i
    character, allocatable :: buffer(:)
    integer :: pos, pack_size
    type(string_t) :: type_name
    integer, parameter :: comm = MPI_COMM_WORLD

    integer :: eq_area_grid_cells = 120
    real(dk) :: eq_area_edges(121), eq_area_midpoints(120), eq_area_deltas(120)

    integer :: csv_grid_cells = 156
    real(dk) :: csv_edges(157), csv_midpoints(156), csv_deltas(156)

    integer :: config_grid_cells = 4
    real(dk) :: config_edges(5), config_deltas(4), config_midpoints(4)

    eq_area_edges = [ 0.000000,1.000000,2.000000,3.000000,4.000000,5.000000,6.000000,7.000000,8.000000,9.000000,&
                      10.00000,11.00000,12.00000,13.00000,14.00000,15.00000,16.00000,17.00000,18.00000,19.00000,&   
                      20.00000,21.00000,22.00000,23.00000,24.00000,25.00000,26.00000,27.00000,28.00000,29.00000,&   
                      30.00000,31.00000,32.00000,33.00000,34.00000,35.00000,36.00000,37.00000,38.00000,39.00000,&   
                      40.00000,41.00000,42.00000,43.00000,44.00000,45.00000,46.00000,47.00000,48.00000,49.00000,&   
                      50.00000,51.00000,52.00000,53.00000,54.00000,55.00000,56.00000,57.00000,58.00000,59.00000,&   
                      60.00000,61.00000,62.00000,63.00000,64.00000,65.00000,66.00000,67.00000,68.00000,69.00000,&   
                      70.00000,71.00000,72.00000,73.00000,74.00000,75.00000,76.00000,77.00000,78.00000,79.00000,&   
                      80.00000,81.00000,82.00000,83.00000,84.00000,85.00000,86.00000,87.00000,88.00000,89.00000,&   
                      90.00000,91.00000,92.00000,93.00000,94.00000,95.00000,96.00000,97.00000,98.00000,99.00000,&   
                      100.0000,101.0000,102.0000,103.0000,104.0000,105.0000,106.0000,107.0000,108.0000,109.0000,&   
                      110.0000,111.0000,112.0000,113.0000,114.0000,115.0000,116.0000,117.0000,118.0000,119.0000,120.0000 ]

    eq_area_midpoints = [ 0.5000000,1.500000,2.500000,3.500000,4.500000,5.500000,6.500000,7.500000,8.500000,9.500000,&
                          10.50000,11.50000,12.50000,13.50000,14.50000,15.50000,16.50000,17.50000,18.50000,19.50000,&   
                          20.50000,21.50000,22.50000,23.50000,24.50000,25.50000,26.50000,27.50000,28.50000,29.50000,&   
                          30.50000,31.50000,32.50000,33.50000,34.50000,35.50000,36.50000,37.50000,38.50000,39.50000,&   
                          40.50000,41.50000,42.50000,43.50000,44.50000,45.50000,46.50000,47.50000,48.50000,49.50000,&   
                          50.50000,51.50000,52.50000,53.50000,54.50000,55.50000,56.50000,57.50000,58.50000,59.50000,&   
                          60.50000,61.50000,62.50000,63.50000,64.50000,65.50000,66.50000,67.50000,68.50000,69.50000,&   
                          70.50000,71.50000,72.50000,73.50000,74.50000,75.50000,76.50000,77.50000,78.50000,79.50000,&   
                          80.50000,81.50000,82.50000,83.50000,84.50000,85.50000,86.50000,87.50000,88.50000,89.50000,&   
                          90.50000,91.50000,92.50000,93.50000,94.50000,95.50000,96.50000,97.50000,98.50000,99.50000,&   
                          100.5000,101.5000,102.5000,103.5000,104.5000,105.5000,106.5000,107.5000,108.5000,109.5000,&   
                          110.5000,111.5000,112.5000,113.5000,114.5000,115.5000,116.5000,117.5000,118.5000,119.5000]

    eq_area_deltas = [ 1.000000,1.000000,1.000000,1.000000,1.000000,1.000000,1.000000,1.000000,1.000000,1.000000,&   
                       1.000000,1.000000,1.000000,1.000000,1.000000,1.000000,1.000000,1.000000,1.000000,1.000000,&   
                       1.000000,1.000000,1.000000,1.000000,1.000000,1.000000,1.000000,1.000000,1.000000,1.000000,&   
                       1.000000,1.000000,1.000000,1.000000,1.000000,1.000000,1.000000,1.000000,1.000000,1.000000,&   
                       1.000000,1.000000,1.000000,1.000000,1.000000,1.000000,1.000000,1.000000,1.000000,1.000000,&   
                       1.000000,1.000000,1.000000,1.000000,1.000000,1.000000,1.000000,1.000000,1.000000,1.000000,&   
                       1.000000,1.000000,1.000000,1.000000,1.000000,1.000000,1.000000,1.000000,1.000000,1.000000,&   
                       1.000000,1.000000,1.000000,1.000000,1.000000,1.000000,1.000000,1.000000,1.000000,1.000000,&   
                       1.000000,1.000000,1.000000,1.000000,1.000000,1.000000,1.000000,1.000000,1.000000,1.000000,&   
                       1.000000,1.000000,1.000000,1.000000,1.000000,1.000000,1.000000,1.000000,1.000000,1.000000,&   
                       1.000000,1.000000,1.000000,1.000000,1.000000,1.000000,1.000000,1.000000,1.000000,1.000000,&   
                       1.000000,1.000000,1.000000,1.000000,1.000000,1.000000,1.000000,1.000000,1.000000,1.000000]

    csv_edges = [ 120.0000,&
                  121.4000,121.9000,122.3000,123.1000,123.8000,124.6000,125.4000,126.2000,127.0000,128.6000,129.4000,130.3000,&
                  132.0000,135.0000,137.0000,145.0000,155.0000,165.0000,170.0000,175.4000,177.0000,178.6000,180.2000,181.8000,&
                  183.5000,185.2000,186.9000,188.7000,190.5000,192.3000,194.2000,196.1000,198.0000,200.0000,202.0000,204.1000,&
                  206.2000,208.3330,210.5260,212.7660,215.0540,217.3910,219.7800,222.2220,224.7190,227.2730,229.8850,232.5580,&
                  235.2940,238.0950,240.9640,243.9020,246.9140,250.0000,253.1650,256.4100,259.7400,263.1580,266.6670,270.2700,&
                  273.9730,277.7780,281.6900,285.7140,289.8550,294.1180,298.5000,302.5000,303.5000,304.5000,305.5000,306.5000,&
                  307.5000,308.5000,309.5000,310.5000,311.5000,312.5000,313.5000,314.5000,317.5000,322.5000,327.5000,332.5000,&
                  337.5000,342.5000,347.5000,352.5000,357.5000,362.5000,367.5000,372.5000,377.5000,382.5000,387.5000,392.5000,&
                  397.5000,402.5000,407.5000,412.5000,417.5000,422.5000,427.5000,432.5000,437.5000,442.5000,447.5000,452.5000,&
                  457.5000,462.5000,467.5000,472.5000,477.5000,482.5000,487.5000,492.5000,497.5000,502.5000,507.5000,512.5000,&
                  517.5000,522.5000,527.5000,532.5000,537.5000,542.5000,547.5000,552.5000,557.5000,562.5000,567.5000,572.5000,&
                  577.5000,582.5000,587.5000,592.5000,597.5000,602.5000,607.5000,612.5000,617.5000,622.5000,627.5000,632.5000,&
                  637.5000,642.5000,647.1000,655.0000,665.0000,675.0000,685.0000,695.0000,705.0000,715.0000,725.0000,735.0000 ]

    csv_midpoints=[120.7000,121.6500,122.1000,122.7000,123.4500,124.2000,125.0000,125.8000,126.6000,127.8000,129.0000,129.8500,&
                   131.1500,133.5000,136.0000,141.0000,150.0000,160.0000,167.5000,172.7000,176.2000,177.8000,179.4000,181.0000,&
                   182.6500,184.3500,186.0500,187.8000,189.6000,191.4000,193.2500,195.1500,197.0500,199.0000,201.0000,203.0500,&
                   205.1500,207.2665,209.4295,211.6460,213.9100,216.2225,218.5855,221.0010,223.4705,225.9960,228.5790,231.2215,&
                   233.9260,236.6945,239.5295,242.4330,245.4080,248.4570,251.5825,254.7875,258.0750,261.4490,264.9125,268.4685,&
                   272.1215,275.8755,279.7340,283.7020,287.7845,291.9865,296.3090,300.5000,303.0000,304.0000,305.0000,306.0000,&
                   307.0000,308.0000,309.0000,310.0000,311.0000,312.0000,313.0000,314.0000,316.0000,320.0000,325.0000,330.0000,&
                   335.0000,340.0000,345.0000,350.0000,355.0000,360.0000,365.0000,370.0000,375.0000,380.0000,385.0000,390.0000,&
                   395.0000,400.0000,405.0000,410.0000,415.0000,420.0000,425.0000,430.0000,435.0000,440.0000,445.0000,450.0000,&
                   455.0000,460.0000,465.0000,470.0000,475.0000,480.0000,485.0000,490.0000,495.0000,500.0000,505.0000,510.0000,&
                   515.0000,520.0000,525.0000,530.0000,535.0000,540.0000,545.0000,550.0000,555.0000,560.0000,565.0000,570.0000,&
                   575.0000,580.0000,585.0000,590.0000,595.0000,600.0000,605.0000,610.0000,615.0000,620.0000,625.0000,630.0000,&
                   635.0000,640.0000,644.8000,651.0500,660.0000,670.0000,680.0000,690.0000,700.0000,710.0000,720.0000,730.0000 ]
    
    csv_deltas = [1.400000,0.500000,0.400000,0.800000,0.700000,0.800000,0.800000,0.800000,0.800000,1.600000,0.800000,0.900000,&
                  1.700000,3.000000,2.000000,8.000000,10.00000,10.00000,5.000000,5.400000,1.600000,1.600000,1.600000,1.600000,&
                  1.700000,1.700000,1.700000,1.800000,1.800000,1.800000,1.900000,1.900000,1.900000,2.000000,2.000000,2.100000,&
                  2.100000,2.133000,2.193000,2.240000,2.288000,2.337000,2.389000,2.442000,2.497000,2.554000,2.612000,2.673000,&
                  2.736000,2.801000,2.869000,2.938000,3.012000,3.086000,3.165000,3.245000,3.330000,3.418000,3.509000,3.603000,&
                  3.703000,3.805000,3.912000,4.024000,4.141000,4.263000,4.382000,4.000000,1.000000,1.000000,1.000000,1.000000,&
                  1.000000,1.000000,1.000000,1.000000,1.000000,1.000000,1.000000,1.000000,3.000000,5.000000,5.000000,5.000000,&
                  5.000000,5.000000,5.000000,5.000000,5.000000,5.000000,5.000000,5.000000,5.000000,5.000000,5.000000,5.000000,&
                  5.000000,5.000000,5.000000,5.000000,5.000000,5.000000,5.000000,5.000000,5.000000,5.000000,5.000000,5.000000,&
                  5.000000,5.000000,5.000000,5.000000,5.000000,5.000000,5.000000,5.000000,5.000000,5.000000,5.000000,5.000000,&
                  5.000000,5.000000,5.000000,5.000000,5.000000,5.000000,5.000000,5.000000,5.000000,5.000000,5.000000,5.000000,&
                  5.000000,5.000000,5.000000,5.000000,5.000000,5.000000,5.000000,5.000000,5.000000,5.000000,5.000000,5.000000,&
                  5.000000,5.000000,4.600000,7.900000,10.00000,10.00000,10.00000,10.00000,10.00000,10.00000,10.00000,10.00000 ]

    config_edges = [ 12.0, 14.0, 16.0, 18.0, 20.0 ]
    config_midpoints = [13.00000,15.00000,17.00000,19.00000]
    config_deltas = [2.0, 2.0, 2.0, 2.0]

    if( musica_mpi_rank( comm ) == 0 ) then
      call grid_tst_config%from_file( config_flsp )
      thewarehouse => grid_warehouse_t( grid_tst_config )
      aGrid => thewarehouse%get_grid( "height", "km" )
      type_name = grid_type_name( aGrid )
      pack_size = thewarehouse%pack_size( comm ) + type_name%pack_size( comm ) + aGrid%pack_size( comm )
      allocate( buffer( pack_size ) )
      pos = 0
      call thewarehouse%mpi_pack( buffer, pos , comm )
      call type_name%mpi_pack(    buffer, pos , comm )
      call aGrid%mpi_pack(        buffer, pos , comm )
      call assert( 319321152, pos <= pack_size )
    end if

    call musica_mpi_bcast( pack_size , comm )
    if( musica_mpi_rank( comm ) .ne. 0 ) allocate( buffer( pack_size ) )
    call musica_mpi_bcast( buffer , comm )

    if( musica_mpi_rank( comm ) .ne. 0 ) then
      pos = 0
      allocate( thewarehouse )
      call thewarehouse%mpi_unpack( buffer, pos , comm )
      call type_name%mpi_unpack(    buffer, pos , comm )
      aGrid => grid_allocate( type_name )
      call aGrid%mpi_unpack(        buffer, pos , comm )
    end if

    call test_grid_t(aGrid, eq_area_grid_cells, eq_area_edges, eq_area_midpoints, eq_area_deltas)
    deallocate( aGrid )

    aGrid => thewarehouse%get_grid( "height", "km" )
    call test_grid_t(aGrid, eq_area_grid_cells, eq_area_edges, eq_area_midpoints, eq_area_deltas)
    deallocate( aGrid )

    aGrid => thewarehouse%get_grid( "wavelength", "nm" )
    call test_grid_t(aGrid, csv_grid_cells, csv_edges, csv_midpoints, csv_deltas)
    deallocate( aGrid )

    aGrid => thewarehouse%get_grid( "time", "hours" )

    call test_grid_t(aGrid, config_grid_cells, config_edges, config_midpoints, config_deltas)
    deallocate( aGrid )

    deallocate( thewarehouse )

  end subroutine test_grids

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

end program test_grid
